
open Util

(*s Datatype of grammars. *)

type terminal = string

type non_terminal = string

type letter = 
  | Terminal of terminal
  | NonTerminal of non_terminal

type production = letter list

type rule = non_terminal * production

module Nt = struct type t = non_terminal let compare = compare end

module Ntset = Set.Make(Nt)
module Ntmap = Map.Make(Nt)

module Tset = Ntset
module Tmap = Ntmap

module R = struct type t = rule let compare = compare end
module Rset = Set.Make(R)

(*s Collection of terminals and non-terminals. *)

let collect start rules = 
  List.fold_left
    (fun (ts,nts) (nt,p) ->
       List.fold_left 
	 (fun (ts,nts) l -> match l with
	    | Terminal t -> (Tset.add t ts, nts)
	    | NonTerminal nt -> (ts, Ntset.add nt nts))
	 (ts, Ntset.add nt nts) p)
    (Tset.empty, Ntset.singleton start)
    rules

(*s Null. *)

type nulls = Ntset.t

let raw_is_null_production nulls p =
  let is_null_letter = function
    | Terminal _ -> false
    | NonTerminal nt -> Ntset.mem nt nulls
  in
  List.for_all is_null_letter p

let null rules = 
  let step nulls =
    List.fold_left 
      (fun ((n,_) as nb) (nt,p) -> 
	 if (not (Ntset.mem nt n)) && raw_is_null_production n p then 
	   (Ntset.add nt n, true)
	 else
	   nb)
      (nulls,false)
      rules
  in
  fixpoint step Ntset.empty

(*s First. *)

type firsts = Tset.t Ntmap.t

let empty_map rules = 
  List.fold_left (fun f (s,_) -> Ntmap.add s Tset.empty f) Ntmap.empty rules

let first_production_step nulls firsts p =
  let rec fold_prod = function
    | [] -> 
	Tset.empty
    | Terminal c :: _ -> 
	Tset.add c Tset.empty
    | NonTerminal nt :: p -> 
	if Ntset.mem nt nulls then 
	  Tset.union (Ntmap.find nt firsts) (fold_prod p)
	else
	  Ntmap.find nt firsts
  in
  fold_prod p

let first rules nulls =
  let step firsts =
    List.fold_left
      (fun (f,b) (nt,p) ->
	 let fnt = Ntmap.find nt f in
	 let fp = first_production_step nulls f p in
	 if Tset.subset fp fnt then 
	   (f,b)
	 else
	   (Ntmap.add nt (Tset.union fnt fp) f, true))
      (firsts,false)
      rules
  in
  fixpoint step (empty_map rules)

(*s Next. *)

type nexts = Tset.t Ntmap.t

let next start rules nulls firsts = 
  let update (nexts,b) nt s =
    let ntn = Ntmap.find nt nexts in
    if Tset.subset s ntn then 
      (nexts,b) 
    else 
      (Ntmap.add nt (Tset.union s ntn) nexts,true)
  in
  let rec update_prod ((nexts,b) as acc) nty = function
    | [] -> 
	acc
    | NonTerminal ntx :: beta -> 
	let acc' = update acc ntx (first_production_step nulls firsts beta) in
	let acc'' = 
	  if raw_is_null_production nulls beta then
	    update acc' ntx (Ntmap.find nty nexts)
	  else
	    acc'
	in
	update_prod acc'' nty beta
    | Terminal _ :: beta -> 
	update_prod acc nty beta
  in
  let step nexts = 
    List.fold_left 
      (fun acc (nt,p) -> update_prod acc nt p) 
      (nexts,false) rules 
  in
  fixpoint step (empty_map rules)

(*s Grammar with pre-computed terminals, non-terminals, nulls, firsts and
    nexts *)

type t = {
  start : non_terminal;
  rules : rule list;
  terminals : Tset.t;
  non_terminals : Ntset.t;
  nulls : nulls;
  firsts : firsts;
  nexts : nexts
}

let create start rules =
  (* we add S# -> S # *)
  let rules = (start ^ "#", [NonTerminal start; Terminal "#"]) :: rules in
  let t,nt = collect start rules in
  let nu = null rules in
  let fi = first rules nu in
  let ne = next start rules nu fi in
  { start = start;
    rules = rules;
    terminals = t;
    non_terminals = nt;
    nulls = nu;
    firsts = fi;
    nexts = ne }

let first_production g = first_production_step g.nulls g.firsts

let is_null_production g = raw_is_null_production g.nulls

(*s Pretty-print *)

open Format

let pp_letter fmt = function
  | Terminal s -> fprintf fmt "\"%s\"" s 
  | NonTerminal s -> fprintf fmt "%s" s

let pp_production fmt = pp_list pp_letter space fmt

let pp_rule fmt (nt,p) = fprintf fmt "%s ::= %a" nt pp_production p

let pp_one_rule fmt curr (s,p) =
  if s <> curr & curr <> "" then fprintf fmt ";@\n";
  if s <> curr then fprintf fmt "%s ::=@\n" s;
  fprintf fmt "  | @[%a@]" pp_production p;
  fprintf fmt "@\n";
  s

let pp fmt g = 
  fprintf fmt "@[";
  ignore (List.fold_left (pp_one_rule fmt) "" g.rules);
  fprintf fmt "@]@."
  
let pp_ntset fmt s = 
  fprintf fmt 
    "@[{ @[%a@] }@]@?" (pp_list pp_print_string comma) (Ntset.elements s)

let pp_tset fmt s = 
  fprintf fmt "@[{ @[%a@] }@]@?"
    (pp_list pp_print_string comma) (Tset.elements s)

let pp_null fmt n = fprintf fmt "Null: %a\n@?" pp_ntset n

let pp_first fmt f =
  Ntmap.iter (fun s f -> fprintf fmt "First(%s) = %a\n" s pp_tset f) f;
  fprintf fmt "@?"

let pp_next fmt ne =
  Ntmap.iter (fun s f -> fprintf fmt "Next(%s) = %a\n" s pp_tset f) ne;
  fprintf fmt "@?"

