
open Util
open Pushdown
open Gram

(*s LR(0) and SLR(1) analyzers. *)

type simple_item = non_terminal * (letter list * letter list)

module Iset = Set.Make(struct type t = simple_item let compare = compare end)

type simple_state = Iset.t

module Sset = Set.Make(Iset)
module Smap = Map.Make(Iset)

type action = simple_state Pushdown.action

module Aset = Set.Make(struct type t = action let compare = compare end)

type lr0_automaton = {
  lr0_start : simple_state;
  lr0_actions : Aset.t Tmap.t Smap.t;
  lr0_gotos : simple_state Ntmap.t Smap.t
}

(*s LR(0) automaton *)

(* closure of a [simple_state] *)

let closure g s =
  let step s = 
    Iset.fold
      (fun i ((s,_) as sb) -> match i with
	 | (_, (_, NonTerminal x :: _)) -> 
	     List.fold_left 
	       (fun ((s,_) as sb) -> function
		 | x', c when x' = x ->
		     let i = x, ([], c) in
		     if Iset.mem i s then sb else (Iset.add i s, true)
		 | _ -> sb)
	       sb
	       g.rules
	 | _ -> sb)
      s
      (s,false)
  in
  fixpoint step s


let lr0_slr1 lr0 g = 
  (* we add S' -> S # *)
  let i' = g.start ^ "#", ([], [NonTerminal g.start; Terminal "#"]) in
  let start = closure g (Iset.singleton i') in
  (* goto/action tables to be filled *)
  let gotos = ref Smap.empty in
  let add_goto s nt s' = 
    gotos := Smap.add s (Ntmap.add nt s' (Smap.find s !gotos)) !gotos
  in
  let actions = ref Smap.empty in
  let add_action s t a =
    let m = Smap.find s !actions in
    let at = try Tmap.find t m with Not_found -> Aset.empty in
    actions := Smap.add s (Tmap.add t (Aset.add a at) m) !actions
  in
  (* closure { [Y -> alpha t . beta] | [Y -> alpha . t beta] in s } *)
  let next s l =
    closure g
      (Iset.fold 
	 (fun (y,w) s' -> match w with
	    | a, l' :: b when l = l' -> Iset.add (y, (l :: a, b)) s'
	    | alpha, _ -> s')
	 s Iset.empty)
  in
  (* rec. construction as a depth-first traversal *)
  let rec build s =
    if not (Smap.mem s !gotos) then begin
      gotos := Smap.add s Ntmap.empty !gotos;
      actions := Smap.add s Tmap.empty !actions;
      Iset.iter
	(fun (y,w) -> match w with
	   | alpha, l :: beta ->
	       let s' = next s l in
	       (match l with
		  | Terminal t ->
		      add_action s t (if t = "#" then Success else Shift s')
		  | NonTerminal nt ->
		      add_goto s nt s');
	       build s'
	   | gamma, [] ->
	       Tset.iter 
		 (fun t -> 
		    if lr0 || Tset.mem t (Ntmap.find y g.nexts) then
		      add_action s t (Reduce (y, List.rev gamma))) 
	       g.terminals)
	s
    end
  in
  build start;
  { lr0_start = start;
    lr0_actions = !actions;
    lr0_gotos = !gotos }

let no_conflict a =
  try
    Smap.iter 
      (fun _ m -> 
	 Tmap.iter (fun _ rs -> if Aset.cardinal rs > 1 then raise Exit) m) 
      a.lr0_actions;
    true
  with Exit -> 
    false

let lr0 = lr0_slr1 true
let slr1 = lr0_slr1 false

let run_lr0 a w = assert false


(*s Pretty-print *)

open Format

let pp_simple_item fmt (s,(l,r)) =
  fprintf fmt "@[%s -> %a . %a@]" s pp_production (List.rev l) pp_production r

let pp_simple_state fmt s =
  fprintf fmt "@[%a@]" (pp_list pp_simple_item newline) (Iset.elements s)

let pp_lr0 fmt a =
  let index =
    let count = ref 0 in
    let table = ref Smap.empty in
    fun s -> 
      try Smap.find s !table 
      with Not_found -> let n = !count in incr count; table := Smap.add s n !table; n
  in
  let pp_action fmt = function
    | Success -> fprintf fmt "success"
    | Shift s -> fprintf fmt "shift %d" (index s)
    | Reduce r -> fprintf fmt "reduce %a" pp_rule r
  in
  let pp_actions fmt a_s = Aset.iter (fun a -> fprintf fmt "%a " pp_action a) a_s in
  Smap.iter
    (fun s acs ->
       let si = index s in
       fprintf fmt "@[state %d:@\n" si;
       fprintf fmt "  @[%a@]@\n@\n" pp_simple_state s;
       fprintf fmt "  @[";
       Tmap.iter
	 (fun t act -> fprintf fmt "%s: @[%a@]@\n" t pp_actions act)
	 acs;
       fprintf fmt "@]@\n  @[";
       Smap.iter
	 (fun s' gs' -> 
	    if Iset.equal s s' then 
	      Ntmap.iter 
		(fun nt s'' -> fprintf fmt "%s: goto %d@\n" nt (index s'')) gs')
	 a.lr0_gotos;
       fprintf fmt "@]@\n@]")
    a.lr0_actions

let analyze_lr0 g =
  printf "LR(0) analyzer:@\n===============@\n";
  let a0 = lr0 g in  
  pp_lr0 std_formatter a0;
  printf "The grammar is %sLR(0)@\n" (if no_conflict a0 then "" else "not ")

let analyze_slr1 g =
  printf "SLR(1) analyzer:@\n===============@\n";
  let a1 = slr1 g in  
  pp_lr0 std_formatter a1;
  printf "The grammar is %sSLR(1)@\n" (if no_conflict a1 then "" else "not ")

